<?PHP
//
// ZendTo
// Copyright (C) 2006 Jeffrey Frey, frey at udel dot edu
// Copyright (C) 2010 Julian Field, Jules at ZendTo dot com 
//
// Based on the original PERL dropbox written by Doke Scott.
// Developed by Julian Field.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//

require_once(NSSDROPBOX_LIB_DIR."NSSDropbox.php");
require_once(NSSDROPBOX_LIB_DIR."NSSUtils.php");
require_once(NSSDROPBOX_LIB_DIR."Timestamp.php");

/*!
  @class NSSDropoff
  
  Wraps an item that's been dropped-off.  There are two methods of
  allocation available.  The primary is using the dropoff ID,
  for which the database will be queried and used to initialize
  the instance.  The second includes no ID, and in this instance
  the $_FILES array will be examined -- if any files were uploaded
  then the dropoff is initialized using $_FILES and $_POST data.
  
  Dropoffs have evolved a bit since the previous version of this
  service.  Each dropoff can now have multiple files associated
  with it, eliminating the need for the end-user to archive
  multiple files for dropoff.  Dropoffs are now created as a
  one-to-many relationship, where the previous version was setup to
  be a one-to-one deal only.
  
  Of course, we're also leveraging the power of SQL to maintain
  the behind-the-scenes data for each dropoff.
*/
class NSSDropoff {

  //  Instance data:
  private $_dropbox = NULL;
  
  private $_dropoffID = -1;
  private $_claimID;
  private $_claimPasscode;
  private $_claimDir;
  
  private $_authorizedUser;
  private $_emailAddr;
  
  private $_auth;
  private $_senderName;
  private $_senderOrganization;
  private $_expiry;
  private $_senderEmail;
  private $_senderIP;
  private $_confirmDelivery;
  private $_informRecipients;
  private $_informPasscode;
  private $_resendWithPasscode;
  private $_note;
  private $_subject;
  private $_created;
  private $_bytes;
  private $_formattedBytes;
  
  private $_recipients;
  
  private $_showIDPasscode = FALSE;
  private $_showPasscodeHTML = TRUE;
  private $_cameFromEmail = FALSE;
  private $_invalidClaimID = FALSE;
  private $_invalidClaimPasscode = FALSE;
  private $_isNewDropoff = FALSE;
  private $_formInitError = NULL;
  private $_okayForDownloads = FALSE;
  
  /*!
    @function dropoffsForCurrentUser
    
    Static function that returns an array of all dropoffs (as
    NSSDropoff instances) that include the currently-authenticated
    user in their recipient list.  
  */
  public static function dropoffsForCurrentUser(
    $aDropbox
  )
  {
    $allDropoffs = NULL;
    
    if ( $targetEmail = strtolower($aDropbox->authorizedUserData('mail')) ) {
      $qResult = $aDropbox->database->DBDropoffsForMe($targetEmail);
      if ( $qResult && ($iMax = count($qResult)) ) {
        //  Allocate all of the wrappers:
        $i = 0;
        while ( $i < $iMax ) {
          $params = $qResult[$i];
          $altParams = array();
          foreach ( $params as $key => $value ) {
            $altParams[preg_replace('/^d\./','',$key)] = $value;
          }
          if ( $nextDropoff = new NSSDropoff($aDropbox, $altParams) ) {
            $allDropoffs[] = $nextDropoff;
          }
          $i++;
        }
      }
    }
    return $allDropoffs;
  }
  
  /*!
    @function dropoffsFromCurrentUser
    
    Static function that returns an array of all dropoffs (as
    NSSDropoff instances) that were created by the currently-
    authenticated user.  Matches are made based on the
    user's username OR the user's email address -- that catches
    authenticated as well as anonymouse dropoffs by the user.
  */
  public static function dropoffsFromCurrentUser(
    $aDropbox
  )
  {
    $allDropoffs = NULL;
    
    if ( $authSender = $aDropbox->authorizedUser() ) {
      $targetEmail = strtolower($aDropbox->authorizedUserData('mail'));
      
      $qResult = $aDropbox->database->DBDropoffsFromMe($authSender, $targetEmail);
      if ( $qResult && ($iMax = count($qResult)) ) {
        //  Allocate all of the wrappers:
        $i = 0;
        while ( $i < $iMax ) {
          if ( $nextDropoff = new NSSDropoff($aDropbox,$qResult[$i]) ) {
            $allDropoffs[] = $nextDropoff;
          }
          $i++;
        }
      }
    }
    return $allDropoffs;
  }

  /*!
    @function dropoffsOutsideRetentionTime
    
    Static function that returns an array of all dropoffs (as
    NSSDropoff instances) that are older than the dropbox's
    retention time.  Subsequently, they should be removed --
    see the "cleanup.php" admin script.
  */
  public static function dropoffsOutsideRetentionTime(
    $aDropbox
  )
  {
    return NSSDropoff::dropoffsOlderThan( $aDropbox,
                              $aDropbox->retainDays() );
  }

  public static function dropoffsNearRetentionTime(
    $aDropbox
  )
  {
    return NSSDropoff::dropoffsOlderThan( $aDropbox,
                              $aDropbox->retainDays() -
                              $aDropbox->warningDays() );
  }

  static function dropoffsOlderThan(
    $aDropbox,
    $days
  )
  {
    $allDropoffs = NULL;
    
    $targetDate = timestampForTime( time() - $days * 24 * 60 * 60 );
    
    $qResult = $aDropbox->database->DBDropoffsTooOld($targetDate);
    if ( $qResult && ($iMax = count($qResult)) ) {
      //  Allocate all of the wrappers:
      $i = 0;
      while ( $i < $iMax ) {
        if ( $nextDropoff = new NSSDropoff($aDropbox,$qResult[$i]) ) {
          $allDropoffs[] = $nextDropoff;
        }
        $i++;
      }
    }
    return $allDropoffs;
  }

  /*!
    @function dropoffsCreatedToday
    
    Static function that returns an array of all dropoffs (as
    NSSDropoff instances) that were made in the last 24 hours.
  */
  public static function dropoffsCreatedToday(
    $aDropbox
  )
  {
    $allDropoffs = NULL;
    
    $targetDate = timestampForTime( time() - 24 * 60 * 60 );
    
    $qResult = $aDropbox->database->DBDropoffsToday($targetDate);
    if ( $qResult && ($iMax = count($qResult)) ) {
      //  Allocate all of the wrappers:
      $i = 0;
      while ( $i < $iMax ) {
        if ( $nextDropoff = new NSSDropoff($aDropbox,$qResult[$i]) ) {
          $allDropoffs[] = $nextDropoff;
        }
        $i++;
      }
    }
    return $allDropoffs;
  }

  /*!
    @function allDropoffs
    
    Static function that returns an array of every single dropoff (as
    NSSDropoff instances) that exist in the database.
  */
  public static function allDropoffs(
    $aDropbox
  )
  {
    $allDropoffs = NULL;
    
    $qResult = $aDropbox->database->DBDropoffsAll();
    if ( $qResult && ($iMax = count($qResult)) ) {
      //  Allocate all of the wrappers:
      $i = 0;
      while ( $i < $iMax ) {
        if ( $nextDropoff = new NSSDropoff($aDropbox,$qResult[$i]) ) {
          $allDropoffs[] = $nextDropoff;
        }
        $i++;
      }
    }
    return $allDropoffs;
  }

  /*!
    @function cleanupOrphans
    
    Static function that looks for orphans:  directories in the dropoff
    directory that have no matching record in the database AND records in
    the database that have no on-disk directory anymore.  Scrubs both
    types of orphans.  This function gets called from the "cleanup.php"
    script after purging "old" dropoffs.
  */
  public static function cleanupOrphans(
    $aDropbox
  )
  {
    $qResult = $aDropbox->database->DBDropoffsAll();
    $scrubCount = 0;
    if ( $qResult && ($iMax = count($qResult)) ) {
      //
      //  Build a list of claim IDs and walk the dropoff directory
      //  to remove any directories that aren't in the database:
      //
      $dropoffDir = $aDropbox->dropboxDirectory();
      if ( $dirRes = opendir($dropoffDir) ) {
        $i = 0;
        $validClaimIDs = array();
        while ( $i < $iMax ) {
          $nextClaim = $qResult[$i]['claimID'];
          
          //  If there's no directory, then we should scrub this entry
          //  from the database:
          if ( !is_dir($dropoffDir."/".$nextClaim) ) {
            if ( $aDropoff = new NSSDropoff($aDropbox,$qResult[$i]) ) {
              $aDropoff->removeDropoff(FALSE);
              echo "- Removed orphaned record:             $nextClaim\n";
            } else {
              echo "- Unable to remove orphaned record:    $nextClaim\n";
            }
            $scrubCount++;
          } else {
            $validClaimIDs[] = $nextClaim;
          }
          $i++;
        }
        while ( $nextDir = readdir($dirRes) ) {
          //  Each item is a NAME, not a PATH.  Test whether it's a directory
          //  and no longer in the database:
          if ( ( $nextDir != '.' && $nextDir != '..' ) && is_dir($dropoffDir."/".$nextDir) && !in_array($nextDir,$validClaimIDs) ) {
            if ( rmdir_r($dropoffDir."/".$nextDir) ) {
              echo "- Removed orphaned directory:          $nextDir\n";
            } else {
              echo "- Unable to remove orphaned directory: $nextDir\n";
            }
            $scrubCount++;
          }
        }
        closedir($dirRes);
      }
    }
    if ( $scrubCount ) {
      printf("%d orphan%s removed.\n\n",$scrubCount,($scrubCount == 1 ? "" : "s"));
    } else {
      echo "No orphans found.\n\n";
    }
  }

  /*!
    @function __construct
    
    Object constructor.  First of all, if we were passed a query result hash
    in $qResult, then initialize the instance using data from the SQL query.
    Otherwise, we need to look at the disposition of the incoming form data:
    
    * The only GET-type form we do comes from the email notifications we
      send to notify recipients.  So the presence of claimID (and possibly
      claimPasscode) in $_GET means we can init as though the user were
      making a pickup.
    
    * If there a POST-type form and a claimID exists in $_POST, then
      try to initialize using that claimID.
    
    * Otherwise, we need to see if the POST-type form data has an action
      of "dropoff" -- if it does, then attempt to create a ~new~ dropoff
      with $_FILES and $_POST.
    
    A _lot_ of state stuff going on in here; might be ripe for simplification
    in the future.
  */
  public function __construct(
    $aDropbox,
    $qResult = FALSE
  )
  {
    $this->_dropbox = $aDropbox;
    
    if ( ! $qResult ) {
      if ( isset($_POST['claimID']) && $_POST['claimID'] ) {
        //  Coming from a web form:
        if ( $this->initWithClaimID($_POST['claimID']) ) {
          $this->_showPasscodeHTML = FALSE;
        } else {
          $this->_invalidClaimID = TRUE;
        }
      } else if ( isset($_GET['claimID']) && $_GET['claimID'] ) {
        //  Coming from an email:
        $this->_cameFromEmail = TRUE;
        if ( ! $this->initWithClaimID($_GET['claimID']) ) {
          $this->_invalidClaimID = TRUE;
        }
      } else if ( isset($_POST['Action']) && $_POST['Action'] == "dropoff" ) {
        $this->_isNewDropoff = TRUE;
        $this->_showPasscodeHTML = FALSE;
        //  Try to create a new one from form data:
        $this->_formInitError = $this->initWithFormData();
        // initWithFormData will set _showIDPasscode if appropriate.
      }
      
      //  If we got a dropoff ID, check the passcode now:
      if ( ! $this->_isNewDropoff && $this->_dropoffID > 0 ) {
        //  Several ways to "authorize" this:
        //
        //    1) if the target user is the currently-logged-in user
        //    2) if the sender is the currently-logged-in user
        //    3) if the incoming form data has the valid passcode
        //
        $curUser = $this->_dropbox->authorizedUser();
        $curUserEmail = $this->_dropbox->authorizedUserData("mail");
        if ( $this->validRecipientEmail($curUserEmail) ||
             ($curUser && ($curUser == $this->_authorizedUser)) ||
             ($curUserEmail && ($curUserEmail == $this->_senderEmail)) ) {
          $this->_showPasscodeHTML = FALSE;
          $this->_okayForDownloads = TRUE;
          $this->_showIDPasscode   = TRUE;
        } else if ( $this->_cameFromEmail ) {
          // If they tried a passcode and it doesn't match, show the HTML
          // but log the failure against them.
          // If their passcode was blank, just show the HTML but don't
          // count the failure.
          $passattempt = trim(@$_GET['claimPasscode']);
          if ($passattempt !== $this->_claimPasscode) {
            // The supplied passcode didn't match, so ask for another
            $this->_showPasscodeHTML = TRUE;
            if ($passattempt == '') {
              // Didn't try, so don't hold it against them.
              $this->_invalidClaimPasscode = FALSE;
            } else {
              // They tried, and failed, so count failure against them.
              $this->_invalidClaimPasscode = TRUE;
            }
          } else {
            $this->_showPasscodeHTML = FALSE;
            $this->_okayForDownloads = TRUE;
            $this->_showIDPasscode   = TRUE;
          }
        } else {
          if ( !$this->_dropbox->authorizedUserData('grantAdminPriv') ) {
            // They aren't an admin, so they must get stuff right
            $passattempt = trim(@$_POST['claimPasscode']);
            if ($passattempt != $this->_claimPasscode) {
              // The supplied passcode didn't match, so ask for another
              $this->_showPasscodeHTML = TRUE;
              if ($passattempt == '') {
                // Didn't try, so don't hold it against them.
                $this->_invalidClaimPasscode = FALSE;
              } else {
                // They tried, and failed, so count failure against them.
                $this->_invalidClaimPasscode = TRUE;
              }
            } else {
              $this->_showPasscodeHTML = FALSE;
              $this->_okayForDownloads = TRUE;
              $this->_showIDPasscode   = TRUE;
            }
          } else {
            $this->_okayForDownloads = TRUE;
            $this->_showIDPasscode   = TRUE;
          }
        }
      }
    } else {
      $this->initWithQueryResult($qResult);
      $this->_showIDPasscode = TRUE;
    }
  }

  /*
    These are all accessors to get the value of all of the dropoff
    parameters.  Note that there are no functions to set these
    parameters' values:  an instance is immutable once it's created!
    
    I won't document each one of them because the names are
    strategically descriptive *grin*
  */
  public function dropbox() { return $this->_dropbox; }
  public function dropoffID() { return $this->_dropoffID; }
  public function claimID() { return $this->_claimID; }
  public function claimPasscode() { return $this->_claimPasscode; }
  public function claimDir() { return $this->_claimDir; }
  public function authorizedUser() { return $this->_authorizedUser; }
  public function auth() { return $this->_auth; }
  public function senderName() { return $this->_senderName; }
  public function senderOrganization() { return $this->_senderOrganization; }
  public function senderEmail() { return $this->_senderEmail; }
  public function expiry() { return $this->_expiry; }
  public function senderIP() { return $this->_senderIP; }
  public function confirmDelivery() { return $this->_confirmDelivery; }
  public function informRecipients() { return $this->_informRecipients; }
  public function informPasscode() { return $this->_informPasscode; }
  public function resendWithPasscode() { return $this->_resendWithPasscode; }
  public function note() { return $this->_note; }
  public function subject() { return $this->_subject; }
  public function created() { return $this->_created; }
  public function recipients() { return $this->_recipients; }
  public function bytes() { return $this->_bytes; }
  public function formattedBytes() { return $this->_formattedBytes; }
  public function formInitError() { return $this->_formInitError; }
  
  /*!
    @function validRecipientEmail
    
    Returns TRUE is the incoming $recipEmail address is a member of the
    recipient list for this dropoff.  Returns FALSE otherwise.
  */
  public function validRecipientEmail(
    $recipEmail
  )
  {
    foreach ( $this->_recipients as $recipient ) {
      if ( strcasecmp($recipient[1],$recipEmail) == 0 ) {
        return TRUE;
      }
    }
    return FALSE;
  }
  
  /*!
    @function files
    
    Returns a hash array containing info for all of the files in
    the dropoff.
  */
  public function files()
  {
    if ( ($dropoffFiles = $this->_dropbox->database->DBFilesByDropoffID($this->_dropoffID)) && (($iMax = count($dropoffFiles)) > 0) ) {
      $fileInfo = array();
      
      $totalBytes = 0.0;
      $i = 0;
      
      while ( $i < $iMax ) {
        $totalBytes += floatval($dropoffFiles[$i++]['lengthInBytes']);
      }
      $dropoffFiles['totalFiles'] = $iMax;
      $dropoffFiles['totalBytes'] = $totalBytes;
      $dropoffFiles['formattedBytes'] = NSSFormattedMemSize($totalBytes);
      return $dropoffFiles;
    }
    return NULL;
  }

  /*!
    @function addFileWithContent
    
    JKF This function is never actually called.

    Add another file to this dropoff's payload, using the provided content,
    filename, and MIME type.
  */
  public function addFileWithContent(
    $content,
    $filename,
    $description,
    $mimeType = 'application/octet-stream'
  )
  {
    if ( ($contentLen = strlen($content)) && strlen($filename) && $this->_dropoffID ) {
      if ( strlen($mimeType) < 1 ) {
        $mimeType = 'application/octet-stream';
      }
      if ( $this->_claimDir ) {
        $tmpname = tempnam($this->_claimDir,'aff_');
        if ( $fptr = fopen($tmpname,'w') ) {
          fwrite($fptr,$content,$contentLen);
          fclose($fptr);
          
          //  Add to database:
          if ( ! $this->_dropbox->database->DBAddFile2( $this->_dropbox, $this->_dropoffID, $tmpname, $filename,
                            $contentLen, $mimeType, $description, $claimID ) ) {
            unlink($tmpname);
            return false;
          }
          return true;
        }
      }
    }
    return false;
  }

  /*!
    @function downloadFile
    
    Given a fileID -- which is simply a rowID from the "file" table in
    the database -- attempt to download that file.  Download requires that
    NO HTTP headers have been transmitted yet, so we have to be very
    careful to call this function BEFORE the PHP has generated ANY output.
    
    We do quite a bit of logging here:
    
    * Log the pickup to the database; this gives the authorized sender
      the ability to examine who made a pick-up, from when and where.
    
    * Log the pickup to the log file -- UID for auth users, 'emailAddr'
      possibly coming in from a form, or anonymously; claim ID; file
      name.
    
    If all goes well, then the user gets a file and we return TRUE.
    Otherwise, and FALSE is returned.
  */
  public function downloadFile(
    $fileID
  )
  {
    global $smarty;

    //  First, make sure we've been properly authenticated:
    if ( $this->_okayForDownloads ) {
      //  Do we have such a file on-record?
      $fileList = $this->_dropbox->database->DBFileList($this->_dropoffID, $fileID);
      if ( $fileList && count($fileList) ) {
        @ob_end_clean(); //turn off output buffering to decrease cpu usage
        header("Content-type: " . htmlspecialchars($fileList[0]['mimeType']));
        // header(sprintf('Content-Disposition: attachment; filename="%s"',
        //     $fileList[0]['basename'])
        //   );
        // If we're talking to IE, encode the attachment filename or else
        // intl characters get garbled completely.
        // Thanks to UxBoD for this.
        // Changed urlencode to rawurlencode on advice from Liam Gretton.
        if (preg_match("/MSIE/", $_SERVER["HTTP_USER_AGENT"])) {
          header('Content-Disposition: attachment; filename="' .
                 rawurlencode($fileList[0]['basename']) . '"');
        } else {
          header('Content-Disposition: attachment; filename="' .
                 $fileList[0]['basename'] . '"');
        }
        header('Content-Transfer-Encoding: binary');

        //  Range-based support stuff:
        header('Last-Modified: ' . substr(gmdate('r', filemtime($this->_claimDir."/".$fileList[0]['tmpname'])), 0, -5) . 'GMT');
        header('ETag: ' . $this->_dropoffID . $fileList[0]['tmpname']);
        header('Accept-Ranges: bytes');
        
        //  No caching, please:
        header('Cache-control: private');
        header('Pragma: private');
        header('Expires: 0');

        // JKF was $fullSize = $fileList[0]['lengthInBytes'];
        // Changed to read real file size not what the DB said.
        // If it was a library file, then the real file might have been
        // changed between the drop-off being made and the file being
        // downloaded. Must have the correct file size.
        $fullSize = filesize($this->_claimDir."/".$fileList[0]['tmpname']);

        //  Multi-thread and resumed downloading should be supported by this next
        //  block:
        if ( isset($_SERVER['HTTP_RANGE']) ) {
          if ( preg_match('/^[Bb][Yy][Tt][Ee][Ss]=([0-9]*)-([0-9]*)$/',$_SERVER['HTTP_RANGE'],$rangePieces) ) {
            if ( is_numeric($rangePieces[1]) && ($offset = intval($rangePieces[1])) ) {
              if ( ($offset >= 0) && ($offset < $fullSize) ) {
                //  Are we doing an honest-to-god range, or a start-to-end range:
                if ( is_numeric($rangePieces[2]) && ($endOfRange = intval($rangePieces[2])) ) {
                  if ( $endOfRange >= 0 ) {
                    if ( $endOfRange >= $fullSize ) {
                      $endOfRange = $fullSize - 1;
                    }
                    if ( $endOfRange >= $offset ) {
                      $length = $endOfRange - $offset + 1;
                    } else {
                      $offset = 0; $length = $fullSize;
                    }
                  } else {
                    $offset = 0; $length = $fullSize;
                  }
                } else {
                  //  start-to-end range:
                  $length = $fullSize - $offset;
                }
              } else {
                $offset = 0; $length = $fullSize;
              }
            } else if ( is_numeric($rangePieces[2]) && ($length = intval($rangePieces[2])) ) {
              //  The last $rangePieces[2] bytes of the file:
              $offset = $fullSize - $length;
              if ( $offset < 0 ) {
                $offset = 0; $length = $fullSize;
              }
            } else {
              $offset = 0;
              $length = $fullSize;
            }
          } else {
            $offset = 0;
            $length = $fullSize;
          }
        } else {
          $offset = 0;
          $length = $fullSize;
        }
        if ( ($offset > 0) && ($length < $fullSize) ) {
          header("HTTP/1.1 206 Partial Content");
          $this->_dropbox->writeToLog(sprintf('Info: downloading claimID %s file %s - partial download of %d bytes, range %d - %d / %d (%s)',
              $this->_claimID,
              $fileList[0]['basename'],
              $length,
              $offset,
              $offset + $length - 1,
              $fullSize,
              $_SERVER['HTTP_RANGE']
            )
          );
        }
        header(sprintf('Content-Range: bytes %d-%d/%d',$offset,$offset + $length - 1,$fullSize));
        header('Content-Length: '.$length);

        //  Open the file:
        ob_start();
        $fptr = fopen($this->_claimDir."/".$fileList[0]['tmpname'],'rb');
        if ($fptr) {
          fseek($fptr,$offset);
          while ( ! feof($fptr) && ! connection_aborted() ) {
            set_time_limit(0);
            print( fread($fptr,8 * 1024) );
            flush();
            ob_flush();
          }
          fclose($fptr);
        }

        //  Who made the pick-up?
        $whoWasIt = $this->_dropbox->authorizedUser();
        $unknown  = gettext('One of the recipients');
        $whoWasItEmail = '';
        $whoWasItUID = '';
        if ( $whoWasIt ) {
          $whoWasItUID = $whoWasIt;
          $whoWasIt = $this->_dropbox->authorizedUserData('displayName');
          $whoWasItEmail = $this->_dropbox->authorizedUserData('mail');
        } else {
          if (isset($_POST['emailAddr']) && strlen($_POST['emailAddr'])>0) {
            $whoWasIt = trim($_POST['emailAddr']);
            $whoWasIt = preg_replace('/[<>]/', '', $whoWasIt); // Protect it
            if (strcasecmp($whoWasIt, $unknown) == 0) {
              $whoWasIt = $unknown;
              $whoWasItEmail = " ";
            } else if (preg_match($this->_dropbox->validEmailRegexp(),
                            $whoWasIt,$wWIParts) ) {
              $whoWasItEmail = $wWIParts[1]."@".$wWIParts[2];
            } else {
              $whoWasIt      = $unknown;
              $whoWasItEmail = " ";
            }
          } else if (strlen($_GET['emailAddr'])>0) {
            $whoWasIt = trim($_GET['emailAddr']);
            $whoWasIt = preg_replace('/[<>]/', '', $whoWasIt); // Protect it
            if (strcasecmp($whoWasIt, $unknown) == 0) {
              $whoWasIt = $unknown;
              $whoWasItEmail = " ";
            } else if (preg_match($this->_dropbox->validEmailRegexp(),
                            $whoWasIt,$wWIParts) ) {
              $whoWasItEmail = $wWIParts[1]."@".$wWIParts[2];
            } else {
              $whoWasIt      = $unknown;
              $whoWasItEmail = " ";
            }
          } else {
            $whoWasIt = $unknown;
            $whoWasItEmail = preg_replace('/ /', '', $whoWasIt);
          }
        }

        //  Only send emails, etc, if the transfer didn't end with an aborted
        //  connection:
        if ( connection_aborted() ) {
          $this->_dropbox->writeToLog(sprintf('Warning: downloading claimID %s file %s - aborted by %s',
                $this->_claimID,
                $fileList[0]['basename'],
                ( $whoWasItUID ? $whoWasItUID : $whoWasIt )
              )
            );
        } else {
          //  Have any pick-ups been made already?
          $extantPickups = $this->_dropbox->database->DBExtantPickups($this->_dropoffID, $whoWasItEmail);
           
          // if ( $this->_confirmDelivery && (! $extantPickups || ($extantPickups[0][0] == 0)) ) {
          if ( $this->_confirmDelivery && ! $extantPickups ) {
            $this->_dropbox->writeToLog("Info: sending pickup confirmation email to ".$this->_senderEmail." for claim ".$this->_claimID);
            $senderHost = gethostbyaddr($this->_senderIP);
            if ($senderHost != '') {
              $senderHost = '(' . $senderHost . ')';
            }
            $smarty->assign('whoWasIt', $whoWasIt);
            $smarty->assign('claimID', $this->_claimID);
            $smarty->assign('filename', htmlentities($fileList[0]['basename'], ENT_NOQUOTES, 'UTF-8'));
            $smarty->assign('remoteAddr', getClientIP());
            $smarty->assign('hostname', gethostbyaddr(getClientIP()));
            $smarty->assign('senderName',     $this->_senderName);
            $smarty->assign('senderOrg',      $this->_senderOrganization);
            $smarty->assign('senderEmail',    $this->_senderEmail);
            $smarty->assign('showIP',         $this->_dropbox->emailSenderIP());
            $smarty->assign('senderIP',       $this->_senderIP);
            $smarty->assign('senderHost',     $senderHost);
            $smarty->assign('createdDate', timeForDate($this->created()));
            $smarty->assign('note',           trim($this->_note));

            $files = $this->files();
            $realFileCount = $files['totalFiles'];
            // Files in email count from 0, files from database count from 0.
            $tplFiles = array();
            for ($i=0; $i<$realFileCount; $i++) {
              $tplFiles[$i] = array();
              $file = $files[$i];
              $tplFiles[$i]['name'] = $file['basename'];
              $tplFiles[$i]['type'] = $file['mimeType'];
              $tplFiles[$i]['size'] = NSSFormattedMemSize($file['lengthInBytes']);
              $tplFiles[$i]['description'] = $file['description'];
              if ($file['checksum'] != '')
                $tplFiles[$i]['checksum'] = $file['checksum'];
              else
                $tplFiles[$i]['checksum'] = gettext('Not calculated');
            }
            $smarty->assignByRef('files', $tplFiles);
            $smarty->assign('fileCount',      $realFileCount);

            $emailSubject = $smarty->getConfigVars('EmailSubjectTag') .
                            sprintf(gettext('%s has picked up your drop-off!'), $whoWasIt);

            if ((preg_match('/^[yYtT1]/', MYZENDTO) && $this->_senderEmail != $whoWasItEmail) || preg_match('/^[^yYtT1]/', MYZENDTO)) {
              $htmltpl = '';
              if ($smarty->templateExists('pickup_email_html.tpl')) {
                $htmltpl = $smarty->fetch('pickup_email_html.tpl');
              }
              if ( ! $this->_dropbox->deliverEmail(
                      $this->_senderEmail,
                      $whoWasItEmail,
                      $emailSubject,
                      $smarty->fetch('pickup_email.tpl'),
                      $htmltpl
                   )
                 ) {
                $this->_dropbox->writeToLog("Error: failed to deliver confirmation email to ".$this->_senderEmail." for claim ".$this->_claimID);
              }
            }
          } else {
            $this->_dropbox->writeToLog("Info: no need to send confirmation email to ".$this->_senderEmail." for claim ".$this->_claimID);
          }
          $this->_dropbox->writeToLog(sprintf("Info: pickup of claimID %s file %s by %s completed.",
                  $this->_claimID,
                  $fileList[0]['basename'],
                  ( $whoWasItUID ? $whoWasItUID : $whoWasIt ))
            );
          
          //  Add to the pickup log:
          $this->_dropbox->database->DBAddToPickupLog($this->_dropbox,
                           $this->_dropoffID,
                           $this->_dropbox->authorizedUser(),
                           $whoWasItEmail,
                           getClientIP(),
                           timestampForTime(time()),
                           $this->_claimID);
        }
        return TRUE;
      }
    }
    return FALSE;
  }

  /*!
    @function resendDropoff
    
    Re-send the dropoff to its recipients.

    If $touchit is 1, then update the datestamp to reset the expiry timer.
    If $touchit is FALSE, then it's a reminder, so change the text.
  */
  public function resendDropoff(
    $touchit = FALSE
  )
  {
        global $NSSDROPBOX_URL;
        global $smarty;

        $senderName = $this->_senderName;
        $senderOrganization = $this->_senderOrganization;
        $senderEmail = $this->_senderEmail;
        // Bug fix by Mark Takeuchi
        // $senderIP = 'unknown IP address';
        // if ( $this->_senderIP ) {
        //   //  Try to get a hostname for the IP, too:
        //   $senderHost = gethostbyaddr($this->_senderIP);
        // }
        $senderIP = $this->_senderIP;
        $senderHost = gethostbyaddr($senderIP);
        if ($senderHost != '') {
          $senderHost = '(' . $senderHost . ')';
        }
        // Note doesn't want to be escaped here as it's escaped in the HTML
        // email template, and we want the plain text too for the plain email.
        // $note = htmlentities($this->_note, ENT_NOQUOTES, 'UTF-8');
        $note = $this->_note;
        $claimID = $this->_claimID;
        $claimPasscode = $this->_claimPasscode;
        $isReminder = FALSE;
        $informPasscode = $this->_resendWithPasscode;
        $files = $this->files();
        $realFileCount = $files['totalFiles'];
        // Work out the real email subject line.
        if ($realFileCount == 1) {
          $emailSubject = sprintf($smarty->getConfigVars(
                                  'EmailSubjectTag') .
                                  gettext('%s has dropped off a file for you'),
                                  $senderName);
        } else {
          $emailSubject = sprintf($smarty->getConfigVars(
                                  'EmailSubjectTag') .
                                  gettext('%s has dropped off files for you'),
                                  $senderName);
        }
        // Files in email count from 0, files from database count from 0.
        $tplFiles = array();
        for ($i=0; $i<$realFileCount; $i++) {
          $tplFiles[$i] = array();
          $file = $files[$i];
          $tplFiles[$i]['name'] = $file['basename'];
          $tplFiles[$i]['type'] = $file['mimeType'];
          $tplFiles[$i]['size'] = NSSFormattedMemSize($file['lengthInBytes']);
          $tplFiles[$i]['description'] = $file['description'];
          if ($file['checksum'] != '')
            $tplFiles[$i]['checksum'] = $file['checksum'];
          else
            $tplFiles[$i]['checksum'] = gettext('Not calculated');
        }

        // Work out the time left if we need to.
        if ($touchit) {
          // Are updating timestamp like we used to.
          $daysLeft    = $this->_dropbox->retainDays();
          $timeLeft    = sprintf(gettext('%d days'), $daysLeft);
          $createdTime = timestampForTime(time());
          $isReminder  = FALSE;
        } else {
          // Not updating timestamp at all
          $secs_gone = time() - timeForDate($this->_created);
          $secs_remaining = $this->_dropbox->retainDays()*86400 - $secs_gone;
          if ($secs_remaining <= 86400) {
            $things_remaining = intval($secs_remaining/3600);
            $things = ($things_remaining==1) ? gettext('1 hour') : gettext('%d hours');
          } else {
            $things_remaining = intval($secs_remaining/86400);
            $things = ($things_remaining==1) ? gettext('1 day') : gettext('%d days');
          }
          $timeLeft = sprintf($things, $things_remaining);
          $daysLeft = intval($secs_remaining/86400); // backwards compat
          $createdTime = timestampForDate($this->_created);
          $isReminder  = TRUE;
        }

        //  Construct the email notification and deliver:
        $smarty->assign('zendToURL',   $NSSDROPBOX_URL);
        $smarty->assign('senderName',  $senderName);
        $smarty->assign('senderOrg',   $senderOrganization);
        $smarty->assign('senderEmail', $senderEmail);
        $smarty->assign('showIP',      $this->_dropbox->emailSenderIP());
        $smarty->assign('senderIP',    $senderIP);
        $smarty->assign('senderHost',  $senderHost);
        $smarty->assign('note',        trim($note));
        $smarty->assign('subject',     $emailSubject);
        $smarty->assign('now',         $createdTime);
        $smarty->assign('claimID',     $claimID);
        $smarty->assign('claimPasscode', $claimPasscode);
        $smarty->assign('informPasscode', $informPasscode);
        $smarty->assign('fileCount',   $realFileCount);
        $smarty->assign('retainDays',  $daysLeft);
        $smarty->assign('timeLeft',    $timeLeft);
        $smarty->assign('isReminder',  $isReminder);
        $smarty->assignByRef('files',  $tplFiles);
        $emailTXTContent = $smarty->fetch('dropoff_email.tpl');

        $emailHTMLContent = '';
        if ($smarty->templateExists('dropoff_email_html.tpl')) {
          $emailHTMLContent = $smarty->fetch('dropoff_email_html.tpl');
        }

        // Make the mail come from the sender, not ZendTo
        foreach ( $this->_recipients as $recipient ) {
          // In MyZendTo, don't send email to myself
          if (preg_match('/^[^yYtT1]/', MYZENDTO) ||
              (preg_match('/^[yYtT1]/', MYZENDTO) && $senderEmail != $recipient[1])) {
            // Bug fix by Sebastian Tyler
            $emailTXTContent2 = preg_replace('/__EMAILADDR__/',
                            urlencode($recipient[1]), $emailTXTContent);
            $emailHTMLContent2 = preg_replace('/__EMAILADDR__/',
                            urlencode($recipient[1]), $emailHTMLContent);
            $success = $this->_dropbox->deliverEmail(
                $recipient[1],
                $senderEmail,
                $emailSubject,
                $emailTXTContent2,
                $emailHTMLContent2
             );
            if ( ! $success ) {
              $this->_dropbox->writeToLog(sprintf("Error: re-delivery of notification email failed to %s for claimID $claimID",$recipient[1]));
            } else {
              // Succeeded, so touch the dropoff's datestamp to reset number
              // of days until auto-deletion
              if ($touchit) {
                $this->_dropbox->database->DBTouchDropoff($claimID,
                                             $createdTime);
              }
              $this->_dropbox->writeToLog(sprintf("Info: re-delivery of notification email succeeded to %s for claimID $claimID",$recipient[1]));
            }
          }
        }

        return TRUE;
  }

  /*!
    @function removeDropoff
    
    Scrub the database and on-disk directory for this dropoff, effectively
    removing it.  We do some writing to the log file to make sure we know
    when this happens.
  */
  public function removeDropoff(
    $doLogEntries = TRUE
  )
  {
    if ( is_dir($this->_claimDir) ) {
      //  Remove the contents of the directory:
      if ( ! rmdir_r($this->_claimDir) ) {
        if ( $doLogEntries ) {
          $this->_dropbox->writeToLog("Error: could not delete drop-off directory ".$this->_claimDir." for claimID ".$this->_claimID);
          $this->_dropbox->writeToLog("Info: drop-off directory for claimID ".$this->_claimID." deleted by ".$this->_dropbox->authorizedUser());
        }
        return FALSE;
      }
    }

    //  Remove any stuff from the database:
    if ( ! $this->_dropbox->database->DBRemoveDropoff($this->_dropbox, $this->_dropoffID, $this->_claimID) ) {
      if ( $doLogEntries ) {
        $this->_dropbox->writeToLog("Info: drop-off records for claimID ".$this->_claimID." removed by ".$this->_dropbox->authorizedUser());
      }
      return FALSE;
    }

    if ( $doLogEntries ) {
      $this->_dropbox->writeToLog("Info: drop-off removal complete for claimID ".$this->_claimID." removed by ".$this->_dropbox->authorizedUser());
    }
    return TRUE;
  }

  /*!
    @function HTMLOnLoadJavascript
    
    Returns the "[form name].[field name]" string that's most appropriate for the page
    that's going to display this object.  Basically allows us to give focus to the
    claim ID or passcode field according to what data we have so far.
  */
  public function HTMLOnLoadJavascript()
  {
    if ( $this->_showPasscodeHTML ) {
      if ( !$this->_invalidClaimID && (@$_GET['claimID'] && !@$_GET['claimPasscode']) || (@$_POST['claimID'] && !@$_POST['claimPasscode']) ) {
        return "pickup.claimPasscode";
      }
      return "pickup.claimID";
    }
    return NULL;
  }

  /*!
    @function HTMLWrite
    
    Composes and writes the HTML that should be output for this
    instance.  If the instance is a fully-initialized, existing
    dropoff, then we'll wind up calling HTMLSummary().  Otherwise,
    we output one of several possible errors (wrong claim passcode,
    e.g.) and possibly show the claim ID and passcode "dialog".
  */
  public function HTMLWrite()
  {
    global $NSSDROPBOX_URL;
    global $smarty;

    $claimID = $this->_claimID;
    
    $smarty->assign('maxBytesForFile', NSSFormattedMemSize($this->_dropbox->maxBytesForFile()));
    $smarty->assign('maxBytesForDropoff', NSSFormattedMemSize($this->_dropbox->maxBytesForDropoff()));
    $smarty->assign('retainDays',  $this->_dropbox->retainDays());

    if ( $this->_invalidClaimID ) {
      NSSError(gettext("The Claim ID or Passcode was incorrect, or the drop-off has expired. Please re-check and note that drop-offs must be collected before they expire otherwise they are automatically deleted."),
               gettext("Invalid Claim ID or Passcode"));
      $claimID = ( $this->_cameFromEmail ? $_GET['claimID'] : $_POST['claimID'] );
    }

    // Sanitise $claimID before using it
    $claimID = preg_replace('/[^a-zA-Z0-9]/', '', trim($claimID));
    if ( $this->_invalidClaimPasscode ) {
      NSSError(gettext("The Claim ID or Passcode was incorrect, or the drop-off has expired. Please re-check and note that drop-offs must be collected before they expire otherwise they are automatically deleted."),
               gettext("Invalid Claim ID or Passcode"));
      // NEW And log the failed attempt at guessing the passcode!
      // Only log and count and delete if it's a real dropoff
      if (! $this->_invalidClaimID) {
        $dID = $this->_dropoffID;
        $this->_dropbox->database->DBAddToPickupLog($this->_dropbox,
                       $dID,
                       $this->_dropbox->authorizedUser(),
                       'FAILEDATTEMPT',
                       getClientIP(),
                       timestampForTime(time()),
                       $claimID);
        // How many failures have we had?
        $failuresCount = $this->_dropbox->database->DBPickupFailures($dID);
        $maxFailures = $this->_dropbox->maxPickupFailures();
        if ($maxFailures>0 && $failuresCount >= $maxFailures) {
          // Too many failures, so this drop-off must self-destruct!
          // Very easy, just delete it from under myself.
          // The rest of this fn will handle the UI just fine.
          $this->_dropbox->writeToLog("Warning: drop-off for claimID ".$claimID." removed due to too many failed pickup attempts (".$failuresCount." >= ".$maxFailures."!");
          $this->removeDropoff(TRUE); // Force logging on
        }
      }
    }
    $smarty->assign('claimID', $claimID);
    if ( $this->_isNewDropoff ) {
      if ( $this->_formInitError ) {
        NSSError($this->_formInitError, gettext("Upload Error"));
      } else {
        $this->HTMLSummary(FALSE,TRUE);
        return 'show_dropoff.tpl';
      }
    }
    else if ( $this->_showPasscodeHTML ) {
      $smarty->assign('cameFromEmail', $this->_cameFromEmail);
      $smarty->assign('emailAddr', $this->_emailAddr);
      return 'claimid_box.tpl';
    } else {
      $this->HTMLSummary(TRUE);
      return 'show_dropoff.tpl';
    }
    return "";
  }

  /*!
    @function HTMLSummary
    
    Compose and write the HTML that shows all of the info for a dropoff.
    This includes:
    
    * A table of claim ID and passcode; sender info; and list of recipients
    
    * A list of the files included in the dropoff.  The icons and names in
      this list will be hyperlinked as download triggers if the $clickable
      argument is TRUE.
    
    * A table of the pickup history for this dropoff.
    
  */
  public function HTMLSummary(
    $clickable = FALSE,
    $overrideShowRecips = FALSE
  )
  {
    global $smarty;

    $curUser = $this->_dropbox->authorizedUser();
    $curUserEmail = $this->_dropbox->authorizedUserData("mail");
    $isSender = FALSE;
    $isAdmin  = FALSE;
    $overrideShowRecips = FALSE;
    if ( $curUser ) {
      if ( $curUserEmail && (strcasecmp($curUserEmail,$this->_senderEmail) == 0) ) {
        $isSender = TRUE;
      }
      if ( $this->_dropbox->authorizedUserData('grantAdminPriv') ) {
        $isAdmin = TRUE;
      }
      if ( ($curUser == $this->_authorizedUser) || $isSender ) {
        $overrideShowRecips = TRUE;
      }
    }
    if ( $this->_senderIP ) {
      //  Try to get a hostname for the IP, too:
      $remoteHostName = gethostbyaddr($this->_senderIP);
    }
    $isSingleRecip = FALSE;
    if ( count($this->_recipients) == 1 ) {
      $isSingleRecip = TRUE;
    }
    $smarty->assign('isClickable', $clickable);

    $smarty->assign('isDeleteable', ( $clickable && ( $isAdmin || $isSender || $isSingleRecip )));
    # Fix, or else you can re-send when emails not allowed.
    # $smarty->assign('isSendable', ( $clickable && ( $isAdmin || $isSender )));
    $smarty->assign('isSendable', ( $clickable && ( $isAdmin || ( $isSender && $this->_dropbox->allowEmailRecipients() ) )));
    $smarty->assign('showIDPasscode', $this->_showIDPasscode);

    $smarty->assign('inPickupPHP', preg_match('/pickup\.php/', $_SERVER['PHP_SELF']));
    $smarty->assign('claimPasscode', $this->_claimPasscode);

    $smarty->assign('senderName',  $this->_senderName);
    $smarty->assign('senderOrg',   htmlspecialchars($this->_senderOrganization));
    $smarty->assign('senderEmail', $this->_senderEmail);
    $smarty->assign('senderHost',  $remoteHostName);
    $smarty->assign('createdDate', timeForDate($this->created()));
    $smarty->assign('expiryDate', timeForDate($this->created())+3600*24*$this->_dropbox->retainDays());
    $smarty->assign('confirmDelivery', $this->_confirmDelivery?"true":"false");
    $smarty->assign('informRecipients', $this->_informRecipients?"true":"false");
    $smarty->assign('allowEmailPasscode', $this->_dropbox->allowEmailPasscode()?"true":"false");
    $smarty->assign('defaultEmailPasscode', $this->_dropbox->defaultEmailPasscode()?"true":"false");

    $smarty->assign('showRecips', ( $this->_dropbox->showRecipsOnPickup() || $overrideShowRecips || ($this->_dropbox->authorizedUser() && $this->_dropbox->authorizedUserData('grantAdminPriv')) ));
    // MyZendTo: If there is only 1 recipient then that must be the sender
    if (preg_match('/^[yYtT1]/', MYZENDTO) && count($this->_recipients)<=1) {
      $smarty->assign('showRecips', FALSE);
    }
    $reciphtml = array();
    foreach($this->_recipients as $r) {
      $reciphtml[] = array(htmlentities($r[0], ENT_NOQUOTES, 'UTF-8'), htmlentities($r[1], ENT_NOQUOTES, 'UTF-8'));
    }
    $smarty->assign('recipients', $reciphtml);

    $smarty->assign('note', htmlentities($this->_note, ENT_NOQUOTES, 'UTF-8'));
    $smarty->assign('subject', htmlentities($this->_subject, ENT_NOQUOTES, 'UTF-8'));

    $dropoffFiles = $this->_dropbox->database->DBFilesForDropoff($this->_dropoffID);
    $smarty->assign('dropoffFilesCount', count($dropoffFiles));

    // Fill the outputFiles array with all the dropoffFiles, over-riding
    // one or two elements as we go so it's ready-formatted.
    $outputFiles = array();
    $i = 0;
    foreach($dropoffFiles as $file) {
      $outputFiles[$i] = $file;
      $outputFiles[$i]['basename'] = htmlentities($file['basename'], ENT_NOQUOTES, 'UTF-8');
      $outputFiles[$i]['length'] = NSSFormattedMemSize($file['lengthInBytes']);
      $outputFiles[$i]['description'] = htmlentities($file['description'],ENT_NOQUOTES, 'UTF-8');
      // Try to find a matching icon from the mimeType
      // Sanitise by changing / -> - and .. -> .
      $mT = str_replace(array('/', '..'), array('-', '.'), $file['mimeType']);
      $mTiconURL = 'images/filetypes/'.strtolower("$mT.svg");
      $mTiconFile = NSSDROPBOX_BASE_DIR.'www/'.$mTiconURL;
      if ($mT !== '' && (is_file($mTiconFile) || is_link($mTiconFile)))
        $outputFiles[$i]['icon'] = $mTiconURL;
      else
        $outputFiles[$i]['icon'] = "images/filetypes/unknown.svg";
      // If we split the checksum string in half, this is where to split it
      $outputFiles[$i]['wrapat'] = floor(strlen($file['checksum'])/2);
      $i++;
    }
    $smarty->assignByRef('files', $outputFiles);

    $emailAddr = isset($this->_emailAddr)?$this->_emailAddr:'';
    if (!$emailAddr) {
      $emailAddr = isset($_POST['emailAddr'])?$_POST['emailAddr']:(isset($_GET['emailAddr'])?$_GET['emailAddr']:NULL);
      $emailAddr = trim($emailAddr);
      if ( strlen($emailAddr)>0 ) {
        if ( preg_match($this->_dropbox->validEmailRegexp(),
                        $emailAddr, $eAParts) )
          $emailAddr = $eAParts[1]."@".$eAParts[2];
        else
          $emailAddr = gettext('One of the recipients');
      } else {
        // No email address at all, so it wasn't invalid but blank.
        // $emailAddr = $smarty->getConfigVars('UnknownRecipient');
        $emailAddr = gettext('One of the recipients');
      }
    }

    $smarty->assign('emailAddr', $emailAddr);
    $smarty->assign('downloadURL', 'download.php?claimID=' . $this->_claimID . '&claimPasscode=' . $this->_claimPasscode . ($emailAddr?('&emailAddr='.urlencode($emailAddr)):''));

    $pickups = $this->_dropbox->database->DBPickupsForDropoff($this->_dropoffID);
    $smarty->assign('pickupsCount', count($pickups));

    // Fill the outputPickups array with all the pickups, over-riding
    // one or two elements as we go so it's ready-formatted.
    $outputPickups = array();
    $i = 0;
    foreach($pickups as $pickup) {
      $outputPickups[$i] = $pickup;
      $hostname = gethostbyaddr($pickups[$i]['recipientIP']);
      if ( $hostname != $pickups[$i]['recipientIP'] ) {
        $hostname = "$hostname (".$pickups[$i]['recipientIP'].")";
      }
      $outputPickups[$i]['hostname'] = htmlentities($hostname, ENT_NOQUOTES, 'UTF-8');
      $outputPickups[$i]['pickupDate'] = timeForTimestamp($pickups[$i]['pickupTimestamp']);
      $authorizedUser = htmlentities($pickups[$i]['authorizedUser'], ENT_NOQUOTES, 'UTF-8');
      if ( ! $authorizedUser ) {
        $authorizedUser = $pickups[$i]['emailAddr'];
      }
      $outputPickups[$i]['pickedUpBy'] = $authorizedUser;
      $i++;
    }
    $smarty->assignByRef('pickups', $outputPickups);
  }

  /*!
    @function ClaimID2Dropoff

    Convert a claimID (random string) into the Dropoff record for it.

    Returns the Dropoff record on success, FALSE otherwise.
  */
  private function ClaimID2Dropoff(
    $claimID
  )
  {
    global $SYSADMIN;
    $claimID = preg_replace('/[^a-zA-Z0-9]/', '', $claimID); // Protect it!
    if ( $this->_dropbox ) {
      $qResult = $this->_dropbox->database->DBDropoffsForClaimID($claimID);
      if ( $qResult && ($iMax = count($qResult)) ) {
        //  Set the fields:
        if ( $iMax == 1 ) {
          return $qResult[0];
        } else {
          NSSError(gettext("There appears to be more than 1 drop-off with that claim ID.").' '.$SYSADMIN, gettext("Invalid Claim ID"));
        }
      }
    }
    return FALSE;
  }

  /*!
    @function initWithClaimID
    
    Completes the initialization (begun by the __construct function)
    by looking-up a dropoff by the $claimID.
    
    Returns TRUE on success, FALSE otherwise.
  */
  private function initWithClaimID(
    $claimID
  )
  {
    $dropoffDBRecord = $this->ClaimID2Dropoff($claimID);
    if (is_array($dropoffDBRecord) &&
        array_key_exists('rowID', $dropoffDBRecord)) {
      $d = $this->initWithQueryResult($dropoffDBRecord);
      // We will end up here if we are doing anything with a drop-off,
      // including re-sending it.
      // If we're doing a pick-up and we haven't yet got a decent email
      // address, pull one from the form data instead. Sanitise!
      if (! $this->_emailAddr && is_array($_POST) &&
          array_key_exists('emailAddr', $_POST)) {
        $addrattempt = trim($_POST['emailAddr']);
        if (preg_match($this->_dropbox->validEmailRegexp(), $addrattempt))
          $this->_emailAddr = $addrattempt;
      }
      // To know whether to resend with the passcode, we need to look at
      // the form data to see if they want us to (if we are allowed to!).
      // Assume we cannot send passcode by default
      $this->_resendWithPasscode = FALSE;
      if ($this->_dropbox->allowEmailPasscode()) {
        // Most people want it set, so start that way
        $this->_resendWithPasscode = TRUE;
        // Get the value of the 'resendWithPasscode' form field
        if (is_array($_POST) &&
            array_key_exists('resendWithPasscode', $_POST) &&
            $_POST['resendWithPasscode'] === "no")
          $this->_resendWithPasscode = FALSE;
        if (is_array($_GET) &&
            array_key_exists('resendWithPasscode', $_GET) &&
            $_GET['resendWithPasscode'] === "no")
          $this->_resendWithPasscode = FALSE;
      }
      return $d;
    }
    return FALSE;
  }
  /*
   * This is what that function used to say before I split it in 2.
    global $SYSADMIN;
    $claimID = preg_replace('/[^a-zA-Z0-9]/', '', $claimID); // Protect it!
    if ( $this->_dropbox ) {
      $qResult = $this->_dropbox->database->DBDropoffsForClaimID($claimID);
      if ( $qResult && ($iMax = count($qResult)) ) {
        //  Set the fields:
        if ( $iMax == 1 ) {
          return $this->initWithQueryResult($qResult[0]);
        } else {
          NSSError(gettext("There appears to be more than 1 drop-off with that claim ID.").' '.$SYSADMIN, gettext("Invalid Claim ID"));
        }
      }
    }
    return FALSE;
  }
  */
  
  /*!
    @function initWithQueryResult
    
    Completes the initialization (begun by the __construct function)
    by pulling instance data from a hash of results from an SQL query.
    
    Also builds an in-memory recipient list by doing a query on the
    recipient table.  The list is a 2D array, each outer element being
    a hash containing values keyed by 'recipName' and 'recipEmail'.
    
    Returns TRUE on success, FALSE otherwise.
  */
  private function initWithQueryResult(
    $qResult
  )
  {
    global $SYSADMIN;
    $trimmed = trim($qResult['claimID']);
    $this->_dropoffID           = $qResult['rowID'];
    
    $this->_claimID             = preg_replace('/[^a-zA-Z0-9]/', '', $qResult['claimID']);
    $this->_claimPasscode       = preg_replace('/[^a-zA-Z0-9]/', '', $qResult['claimPasscode']);
    
    $this->_authorizedUser      = $qResult['authorizedUser'];
    $this->_emailAddr           = @$qResult['emailAddr'];
    
    $this->_senderName          = $qResult['senderName'];
    $this->_senderOrganization  = $qResult['senderOrganization'];
    $this->_senderEmail         = $qResult['senderEmail'];
    $this->_note                = @$qResult['note'];
    $this->_subject             = @$qResult['subject'];

    $this->_senderIP            = $qResult['senderIP'];
    $this->_confirmDelivery     = ( isset($qResult['confirmDelivery']) && preg_match('/[tT1]/',$qResult['confirmDelivery']) ) ? TRUE : FALSE;
    $this->_informRecipients    = ( isset($qResult['informRecipients']) && preg_match('/[tT1]/',$qResult['informRecipients']) ) ? TRUE : FALSE;
    $this->_informPasscode      = ( isset($qResult['informPasscode']) && preg_match('/[tT1]/',$qResult['informPasscode']) ) ? TRUE : FALSE;
    $this->_created             = dateForTimestamp($qResult['created']);
    
    $this->_recipients          = $this->_dropbox->database->DBRecipientsForDropoff($qResult['rowID']);
    $this->_bytes               = $this->_dropbox->database->DBBytesOfDropoff($qResult['rowID']);
    $this->_formattedBytes      = NSSFormattedMemSize($this->_bytes);
    
    // Verify emailAddr is legal and not malicious.
    // Must be non-blank and must match our regexp.
    // If malicious, just blank it.
    if ( isset($this->_emailAddr) && trim($this->_emailAddr) !== '' &&
	 ! preg_match($this->_dropbox->validEmailRegexp(), $this->_emailAddr)) {
      $this->_emailAddr = '';
    }

    if ( ! $this->_dropbox->directoryForDropoff($trimmed, $this->_claimDir) ) {
      NSSError(gettext("The directory containing this drop-off's files has gone missing.").' '.$SYSADMIN, gettext("Drop-off Directory Not Found"));
      return FALSE;
    } else {
      return TRUE;
    }
  }
  
  // Work out how many files they submitted in the form.
  private $maxFilesKey = 0; // This holds the biggest file index we ever look at
  private function numberOfFiles() {
    $i=1;
    $files=0;
    while ($i<200) { // Okay, we can never have more than 200 files in 1 dropoff
      if ((array_key_exists("file_select_".$i, $_POST) &&
           $_POST["file_select_".$i] != -1) ||
          (array_key_exists("file_".$i, $_FILES) &&
           array_key_exists("tmp_name", $_FILES["file_".$i]) &&
           $_FILES["file_".$i]['tmp_name'] !== "")) {
        $files++;
        $this->maxFilesKey = $i;
      }
      $i++;
    }
    return $files;
  }



  /*!
    @function initWithFormData
    
    This monster routine examines POST-type form data coming from our dropoff
    form, validates all of it, and actually creates a new dropoff.
    
    The validation is done primarily on the email addresses that are involved,
    and all of that is documented inline below.  We also have to be sure that
    the user didn't leave any crucial fields blank.
    
    We examine the incoming files to be sure that individually they are all
    below our parent dropbox's filesize limit; in the process, we sum the
    sizes so that we can confirm that the whole dropoff is below the parent's
    dropoff size limit.
    
    Barring any problems with all of that, we get a new claimID and claim
    directory for this dropoff and move the uploaded files into it.  We add
    a record to the "dropoff" table in the database.
    
    We also have to craft and email and send it to all of the recipients.  A
    template string is created with the content and then filled-in individually
    (think form letter) for each recipient (we embed the recipient's email address
    in the URL so that it _might_ be possible to identify the picker-upper even
    when the user isn't logged in).
    
    If any errors occur, this function will return an error string.  But
    if all goes according to plan, then we return NULL!
  */
  private function initWithFormData()
  {
    global $NSSDROPBOX_URL;
    global $smarty;
    global $BACKBUTTON;
    global $SYSADMIN;
    
    // Start off with the data from the form posting, overwriting it with
    // stored data as necessary.
    $senderName = paramPrepare(@$_POST['senderName']);
    $senderName = preg_replace('/[<>]/', '', $senderName);
    $senderEmail = paramPrepare(strtolower(@$_POST['senderEmail']));
    $senderOrganization = paramPrepare($_POST['senderOrganization']);
    $senderOrganization = preg_replace('/[<>]/', '', $senderOrganization);
    $note = $_POST['note'];
    $expiry = 0;

    // Assume they are an outsider until proven otherwise.
    // Outsiders don't get to see the ClaimID or Passcode of their
    // drop-off, so they can't send it to other outsiders.
    $showIDPasscode = FALSE;

    // If they have a valid req key, then they don't need to be verified
    // or logged in.
    $reqSubject = '';
    $req = '';
    if ($_POST['req'] != '') {
      $dummy = '';
      $recipName = '';  // Never actually use this
      $recipEmail = ''; // Never actually use this
      $req = preg_replace('/[^a-zA-Z0-9]/', '', $_POST['req']);
      if ($this->_dropbox->ReadReqData($req,
                                       $recipName, $recipEmail,
                                       $senderOrganization,
                                       $senderName, $senderEmail,
                                       $dummy, $reqSubject, $expiry)) {
        if ($expiry < time()) {
          $this->_dropbox->DeleteReqData($req);
          return gettext("Failed to read your verification information. Your drop-off session has probably expired. Please start again.");
        }
        // It was a valid req key, so leave $req alone (and true).
        $reqSubject = trim($reqSubject);
        $this->_subject = $reqSubject;
      } else {
        // Invalid request code, so ignore them
        $req = FALSE;
        $reqSubject = '';
      }
    }

    // It is not a request, or not a valid request
    if ($req == '') {
      // So now they must be authorized as it's not a request
      if (! $this->_dropbox->authorizedUser()) {
        $auth = $_POST['auth'];

        // JKF Get the above from the auth database
        // JKF Fail if it doesn't exist or it's a pickup auth not a dropoff
        $authdatares = $this->_dropbox->ReadAuthData($auth,
                                             $senderName,
                                             $senderEmail,
                                             $senderOrganization,
                                             $expiry);
        if (! $authdatares) {
          return gettext("Failed to read your verification information. Your drop-off session has probably expired. Please start again.");
        }
        // If the email is blank (and name has no spaces) then it's a pickup.
        // In a pickup, the name is used as the sender's IP address.
        if (!preg_match('/ /', $senderName) && $senderEmail == '') {
          return gettext("Failed to read your verification information. Your drop-off session has probably expired. Please start again.");
        }
        if ($expiry < time()) {
          $this->_dropbox->DeleteAuthData($auth);
          return gettext("Failed to read your verification information. Your drop-off session has probably expired. Please start again.");
        }
      } else {
        // Logged-in user so just read their data
        $senderName = $this->_dropbox->authorizedUserData("displayName");
        $senderOrganization = paramPrepare(@$_POST['senderOrganization']);
        $senderOrganization = preg_replace('/[<>]/', '', $senderOrganization);
        $senderEmail = trim($this->_dropbox->authorizedUserData("mail"));
        // Remember they are a member of our organisation,
        // needed when constructing page showing contents of drop-off.
        $showIDPasscode = TRUE;
      }
    }


    // Erase the note if it is just white space.
    if (preg_match('/^\s*$/', $note)) {
      $note = "";
    }

    // Check the length of the note.
    // Don't reject it any more, just trim it if it's too long.
    if (! function_exists('mb_strimwidth')) {
      return gettext('Admin: Install the PHP module/rpm "mbstring".');
    }
    $maxlen = $this->_dropbox->maxnotelength();
    $note = mb_strimwidth($note, 0, $maxlen, '...');

    $confirmDelivery = ( @$_POST['confirmDelivery'] ? TRUE : FALSE );
    $informRecipients = ( @$_POST['informRecipients'] ? TRUE : FALSE );
    $informPasscode = ( @$_POST['informPasscode'] ? TRUE : FALSE );
    $checksumFiles = ( @$_POST['checksumFiles'] ? TRUE : FALSE );
    
    $recipients = array();
    $recipIndex = -1; // <0 => no recipients found
    $arraykeys = array_keys($_POST);
    foreach ($arraykeys as $arraykey) {
     $matches = array();
     if ( preg_match('/^recipient_(\d+)/', $arraykey, $matches) ) {
      $recipIndex = $matches[1];
      //while ( array_key_exists('recipient_'.$recipIndex,$_POST) ) {
      $recipName = paramPrepare($_POST['recipName_'.$recipIndex]);
      $recipName = mb_strimwidth($recipName, 0, 100, '...');
      $recipEmail = paramPrepare($_POST['recipEmail_'.$recipIndex]);
      if ( $recipName || $recipEmail ) {
        //  Take the email to purely lowercase for simplicity:
        $recipEmail = strtolower($recipEmail);
         
        //  Just a username?  We add an implicit "@domain.com" for these and validate them!
        $emailParts[1] = NULL;
        $emailParts[2] = NULL;
        if ( preg_match('/\@/',$recipEmail) ) {
          // Has an @ sign so is an email address. Must be valid!
          if ( ! preg_match($this->_dropbox->validEmailRegexp(),$recipEmail,$emailParts) ) {
            return sprintf(gettext("The recipient email address '%s' is invalid."), htmlspecialchars($recipEmail)).' '.$BACKBUTTON;
           }
        } else {
          // No @ sign so just stick default domain in right hand side
          $emailParts[1] = $recipEmail;
          $emailParts[2] = $this->_dropbox->defaultEmailDomain();
        }
        $recipEmailDomain = $emailParts[2];
        // Don't think this line is needed any more, but harmless
        $recipEmail = $emailParts[1]."@".$emailParts[2];
    
        //  Look at the recipient's email domain; un-authenticated users can only deliver
        //  to the dropbox's domain:
        // JKF Changed checkRecipientDomain to return true if it's a local user
        if ( ! $this->_dropbox->authorizedUser() && ! $this->_dropbox->checkRecipientDomain($recipEmail) ) {
                // TRANSLATORS: %1$s = OrganizationShortName
                return sprintf(gettext('You must be logged in as a %1$s user in order to drop-off a file for a non-%1$s user.'), $smarty->getConfigVars('OrganizationShortName')) .
                       '<br/>&nbsp;<br/>' .
                       sprintf(gettext("Return to the %s main menu to log in and then try again."), $smarty->getConfigVars('ServiceTitle'));
        }
        $recipients[] = array(( $recipName ? $recipName : "" ),$recipEmail);
      } else if ( $recipName && !$recipEmail ) {
        return gettext("You must specify all recipients' email addresses in the form.").' '.$BACKBUTTON;
      }
      //$recipIndex++;
     }
    }
    // No recipients found?
    if ( $recipIndex<0 ) {
      return gettext("You must specify all recipients' email addresses in the form.").' '.$BACKBUTTON;
    }

    
    //
    //  Check for an uploaded CSV/TXT file containing addresses:
    //
    if ( isset($_FILES) && isset($_FILES['recipient_csv']) && $_FILES['recipient_csv']['tmp_name'] ) {
      if ( $_FILES['recipient_csv']['error'] != UPLOAD_ERR_OK ) {
        // TRANSLATORS: %s = filename
        $error = sprintf(gettext("There was an error while uploading '%s'."), htmlspecialchars($_FILES['recipient_csv']['name']));
        switch ( $_FILES['recipient_csv']['error'] ) {
          case UPLOAD_ERR_INI_SIZE:
            $error .= gettext("The recipients file size exceeds the limit imposed by PHP on the server.").' '.$SYSADMIN;
            break;
          case UPLOAD_ERR_FORM_SIZE:
            $error .= sprintf(gettext("The recipients file size exceeds the size limit (the maximum is %s)."), NSSFormattedMemSize($this->_dropbox->maxBytesForFile()));
            break;
          case UPLOAD_ERR_PARTIAL:
            $error .= gettext("The recipients file was only partially uploaded. Your network connection may have timed out while attempting to upload.");
            break;
          case UPLOAD_ERR_NO_FILE:
            $error .= gettext("No recipient file was actually uploaded.");
            break;
          case UPLOAD_ERR_NO_TMP_DIR:
            $error .= gettext("The server was not configured with a temporary folder for uploads.").' '.$SYSADMIN;
            break;
          case UPLOAD_ERR_CANT_WRITE:
            $error .= gettext("The server's temporary folder is misconfigured.").' '.$SYSADMIN;
            break;
        }
        return $error;
      }
      
      //  Parse the CSV/TXT file:
      if ( $csv = fopen($_FILES['recipient_csv']['tmp_name'],'r') ) {
        while ( $fields = fgetcsv($csv) ) {
          if ( $fields[0] !== NULL ) {
            //  Got one; figure out which field is an email address:
            foreach ( $fields as $recipEmail ) {
              //  Take the email to purely lowercase for simplicity:
              $recipEmail = strtolower($recipEmail);
               
              // JKF Don't allow just usernames in CSV file!
              if ( ! preg_match($this->_dropbox->validEmailRegexp(),$recipEmail,$emailParts) ) {
                continue;
              }
              $recipEmailDomain = $emailParts[2];
              $recipEmail = $emailParts[1]."@".$emailParts[2];
          
              //  Look at the recipient's email domain;
              //  un-authenticated users can only deliver to the dropbox's
              //  domain:
              if ( ! $this->_dropbox->authorizedUser() && (! $this->_dropbox->checkRecipientDomain($recipEmail)) ) {
                // TRANSLATORS: %1$s = OrganizationShortName
                return sprintf(gettext("You must be logged in as a %1$s user in order to drop-off a file for a non-%$1s user."), $smarty->getConfigVars('OrganizationShortName')) .
                       '<br/>&nbsp;<br/>' .
                       sprintf(gettext("Return to the %s main menu to log in and then try again."), $smarty->getConfigVars('ServiceTitle'));
              }
              // $recipients[] = array(( $recipName ? $recipName : "" ),$recipEmail);
              $recipients[] = array("",$recipEmail);
            }
          }
        }
        fclose($csv);
      } else {
        return gettext("Could not read the uploaded recipients file.");
      }
    }
    
    // If it's in response to a request, and the recipient override
    // is set, then zap the first recipient and replace with ours.
    $reqRecipient = $this->_dropbox->reqRecipient();
    if ($req != '' && $reqRecipient != '') {
      $recipients[0][1] = $reqRecipient;
    }

    // Reduce the list of recipients to those with unique email addresses
    uniqueifyRecipients($recipients);

    //  Confirm that all fields are present and accounted for:
    $fileCount = $this->numberOfFiles();
    if ( $fileCount == 0 ) {
      return gettext("You must choose at least one file to drop-off.").' '.$BACKBUTTON;
    }
    
    //  Now make sure each file was uploaded successfully, isn't too large,
    //  and that the total size of the upload isn't over capacity:
    $i = 1;
    $totalBytes = 0.0;
    $totalFiles = 0;
    // while ( $i <= $fileCount ) {
    while ( $i <= $this->maxFilesKey ) {
      $key = "file_".$i;
      if ( array_key_exists('file_select_'.$i, $_POST) && $_POST['file_select_'.$i] != "-1" ) {
        $totalFiles++;
      } elseif ( $_FILES[$key]['name'] ) {
        if ( $_FILES[$key]['error'] != UPLOAD_ERR_OK ) {
          // TRANSLATORS: %s = filename
          $error = sprintf(gettext("There was an error while uploading '%s'."), htmlspecialchars($_FILES[$key]['name']));
          switch ( $_FILES[$key]['error'] ) {
            case UPLOAD_ERR_INI_SIZE:
              $error .= gettext("The file size exceeds the limit imposed by PHP on the server.").' '.$SYSADMIN;
              break;
            case UPLOAD_ERR_FORM_SIZE:
              $error .= sprintf(gettext("The file '%s' was too large. Each dropped-off file may be at most %s."), htmlspecialchars($_FILES[$key]['name']), NSSFormattedMemSize($this->_dropbox->maxBytesForFile()));
              break;
            case UPLOAD_ERR_PARTIAL:
              $error .= gettext("The file was only partially uploaded. Your network connection may have timed out while attempting to upload.");
              break;
            case UPLOAD_ERR_NO_FILE:
              $error .= gettext("No file was actually dropped-off.");
              break;
            case UPLOAD_ERR_NO_TMP_DIR:
              $error .= gettext("The server was not configured with a temporary folder for uploads.").' '.$SYSADMIN;
              break;
            case UPLOAD_ERR_CANT_WRITE:
              $error .= gettext("The server's temporary folder is misconfigured.").' '.$SYSADMIN;
              break;
          }
          return $error;
        }
        if ( ($bytes = $_FILES[$key]['size']) < 0 ) {
          //  Grrr...stupid 32-bit nonsense.  Convert to the positive
          //  value float-wise:
          $bytes = ($bytes & 0x7FFFFFFF) + 2147483648.0;
        }
        if ( $bytes > $this->_dropbox->maxBytesForFile() ) {
          return sprintf(gettext("The file '%s' was too large. Each dropped-off file may be at most %s."), htmlspecialchars($_FILES[$key]['name']), NSSFormattedMemSize($this->_dropbox->maxBytesForFile()));
        }
        if ( ($totalBytes += $bytes) > $this->_dropbox->maxBytesForDropoff() ) {
          return sprintf(gettext("The total size of the dropped-off files exceeds the maximum for a single drop-off. Altogether, a single drop-off can be at most %s."),
                      NSSFormattedMemSize($this->_dropbox->maxBytesForDropoff())
                    );
        }
        // MyZendTo: If they don't have enough quota left, disallow it
        if (preg_match('/^[yYtT1]/', MYZENDTO)) {
          $QuotaLeft = $this->_dropbox->database->DBRemainingQuota($this->_dropbox->authorizedUser());
          if ( $totalBytes > $QuotaLeft ) {
            return sprintf(gettext("You do not have enough remaining storage quota for this drop-off. You need at least another %s."), NSSFormattedMemSize($totalBytes-$QuotaLeft));
          }
        }
        $totalFiles++;
      }
      $i++;
    }
    if ( $totalFiles == 0 ) {
      return gettext("You must choose at least one file to drop-off.").' '.$BACKBUTTON;
    }

    // Call clamdscan on all the files, fail if they are infected
    // If the name of the scanner is set to '' or 'DISABLED' then skip this.
    $clamdscancmd = $this->_dropbox->clamdscan();
    if ($clamdscancmd != 'DISABLED') {
      $ccfilecount = 1;
      $ccfilelist = '';
      $foundsometoscan = FALSE;
      while ( $ccfilecount <= $this->maxFilesKey ) {
        // For every possible file, we only add it to the list of clamd
        // targets if it's not a library file (assumed clean), but it is
        // an uploaded file, and that file-slot in the form was used.
        $filekey = "file_".$ccfilecount;
        $selectkey = "file_select_".$ccfilecount;
        if (!(array_key_exists($selectkey, $_POST) && // If not library file
              $_POST[$selectkey] != "-1") &&
            array_key_exists($filekey, $_FILES) &&   // and is uploaded file
            array_key_exists('tmp_name', $_FILES[$filekey]) &&
            $_FILES[$filekey]['tmp_name'] !== '') {  // and is not blank
          $ccfilelist .= ' ' . $_FILES[$filekey]['tmp_name'];
          $foundsometoscan = TRUE;
        }
        $ccfilecount++;
      }
      if ($foundsometoscan) { // Don't do any of this if they uploaded nothing
        exec("/bin/chmod go+r " . $ccfilelist); // Need clamd to read them!
        $clamdinfected = 0;
        $clamdoutput = array();
        $clamcmd = exec($clamdscancmd . $ccfilelist,
                        $clamdoutput, $clamdinfected);
        if ($clamdinfected == 1) {
          # JKF 2017-02-27 Added more logging
          $this->_dropbox->writeToLog("Warning: Virus scan of dropped-off files ".$ccfilelist." for ".$this->_dropbox->authorizedUser()." found virus ".implode(' ', $clamdoutput));
          return gettext("One or more of the files you dropped-off was infected with a virus. The drop-off has been abandoned. Please clean your files and try again.");
        }
        if ($clamdinfected == 2) {
          # JKF 2017-02-27 Added more logging
          $this->_dropbox->writeToLog("Error: Virus scan of dropped-off files ".$ccfilelist." for ".$this->_dropbox->authorizedUser()." failed with ".implode(' ', $clamdoutput));
          return gettext("The attempt to virus-scan your drop-off failed.").' '.$SYSADMIN;
        }
      }
    }

    // Call sha256sum on all the files, if they wanted checksums and if
    // the total drop-off size is below our limit (or else it takes too long!)
    // Do all the checks to confirm whether we are actually going to
    // checksum the files or not. From now on we just check $checksumFiles
    // which will be true or false.
    $checksumcmd = $this->_dropbox->checksum();
    if ( $checksumcmd === 'DISABLED' || $checksumcmd === '' ||
         $totalBytes > $this->_dropbox->maxBytesForChecksum()) {
      $checksumFiles = FALSE;
    }

    if ( ! $senderName ) {
      return gettext("You must specify your name in the form.").' '.$BACKBUTTON;
    }
    if ( ! $senderEmail ) {
      return gettext("You must specify your own email address in the form.").' '.$BACKBUTTON;
    }
    if ( ! preg_match($this->_dropbox->validEmailRegexp(),$senderEmail,$emailParts) ) {
      return gettext("The sender email address you entered was invalid.").' '.$BACKBUTTON;
    }
    $senderEmail = $emailParts[1]."@".$emailParts[2];
    
    //  Invent a passcode and claim ID:
    $claimPasscode = NSSGenerateCode();
    $claimID = NULL; $claimDir = NULL;
    if ( ! $this->_dropbox->directoryForDropoff($claimID,$claimDir) ) {
      return gettext("A unique directory to contain your dropped-off files could not be created.").' '.$SYSADMIN;
    }
    
    //  Insert into database:
    if ( $this->_dropbox->database->DBStartTran() ) {
      if ( $dropoffID = $this->_dropbox->database->DBAddDropoff($claimID,
                          $claimPasscode,
                          $this->_dropbox->authorizedUser(),
                          $senderName, $senderOrganization, $senderEmail,
                          getClientIP(),
                          $confirmDelivery,
                          timestampForTime(time()),
                          $note) ) {

        //  Add recipients:
        if ( ! $this->_dropbox->database->DBAddRecipients($recipients, $dropoffID) ) {
          $this->_dropbox->database->DBRollbackTran();
          return gettext("Could not add recipients to the database.").' '.$SYSADMIN;
        }
        
        //  Process the files:
        $i = 1;
        $realFileCount = 0;
        $tplFiles = array(); // These are the file hashes we process in tpl.
        //while ( $i <= $fileCount ) {
        while ( $i <= $this->maxFilesKey ) {
          $key = "file_".$i;
          $selectkey = 'file_select_'.$i;
          if ( array_key_exists($selectkey, $_POST) &&
               $_POST[$selectkey] != "-1" ) {
            // It's a library file.
            // Get the name of the library file they want (safely)
            // by removing all "../" elements and things like it
            $libraryfile = preg_replace('/\.\.[:\/\\\]/', '', $_POST[$selectkey]);
            $libraryfile = preg_replace('/\</', '', $libraryfile); // Protect further
            $libraryfile = paramPrepare($libraryfile);
            // Generate a random filename (collisions are very unlikely)
            $tmpname = mt_rand(10000000, 99999999);
            // Link in the library file
            symlink($this->_dropbox->libraryDirectory().'/'.$libraryfile,
                    $claimDir.'/'.$tmpname);

            // Now strip off the possible subdirectory name as we only
            // want it in the symlink and not after that.
            $libraryfilesize = filesize($this->_dropbox->libraryDirectory().'/'.$libraryfile);
            $libraryfile = trim(preg_replace('/^.*\//', '', $libraryfile));

            // We use this a few times
            $librarydesc = paramPrepare(trim($_POST["desc_".$i]));
            $librarydesc = mb_strimwidth($librarydesc, 0, 100, '...');

            // Checksum it if small enough (awkward to check size)
            $checksum = '';
            if ($checksumFiles &&
                $libraryfilesize <= $this->_dropbox->maxBytesForChecksum()) {
              $checksum = checksumFile($checksumcmd, $claimDir.'/'.$tmpname);
            }

            //  Add to database:
            if ( ! $this->_dropbox->database->DBAddFile1($dropoffID, $tmpname,
                             $libraryfile,
                             $libraryfilesize,
                             "application/octet-stream",
                             $librarydesc,
                             $checksum) ) {
              //  Exit gracefully -- dump database changes and remove the dropoff
              //  directory:
              $this->_dropbox->writeToLog("Error: failed to add dropoff library file $libraryfile to database for $claimID, backing out new dropoff");
              if ( ! rmdir_r($claimDir) ) {
                $this->_dropbox->writeToLog("Error: backing out new dropoff, unable to remove $claimDir - orphaned!");
              }
              if ( ! $this->_dropbox->database->DBRollbackTran() ) {
                $this->_dropbox->writeToLog("Error: backing out new dropoff, failed to ROLLBACK after botched dropoff $claimID, there may be orphan files");
              }
              // TRANSLATORS: %s = filename (of library file)
              return sprintf(gettext("Trouble while attempting to save the information for '%s'."), $libraryfile).' '.$SYSADMIN;
            }
            
            //  That's right, one more file!
            $tplFiles[$realFileCount] = array();
            $tplFiles[$realFileCount]['name'] = $libraryfile;
            $tplFiles[$realFileCount]['type'] = gettext('Library');
            $tplFiles[$realFileCount]['size'] = NSSFormattedMemSize($libraryfilesize);
            $tplFiles[$realFileCount]['description'] = $librarydesc;
            if ($checksum != '')
              $tplFiles[$realFileCount]['checksum'] = $checksum;
            else
              $tplFiles[$realFileCount]['checksum'] = gettext('Not calculated');
            $realFileCount++;
            
            // Update the description in the library index
            $this->_dropbox->database()->DBUpdateLibraryDescription($libraryfile, $librarydesc);
          } elseif ( $_FILES[$key]['name'] ) {
            // It's an uploaded file
            $tmpname = basename($_FILES[$key]['tmp_name']);
            // Get file size from local copy, not what browser told us
            $bytes = filesize($_FILES[$key]['tmp_name']);
            if ( $tmpname == "" || $bytes < 1 ) {
              // I think we've got a screwed upload.
              // A file with no temporary name and no bytes in it?
              // Really?
              // Find the free space for new uploads
              $freeSpace = disk_free_space(ini_get('upload_tmp_dir'));
              // If the space check worked & there's less than 1 upload space
              if ($freeSpace !== false && $freespace <= $this->_dropbox->maxBytesForDropoff())
                return sprintf(gettext("It looks like your drop-off failed due to lack of free space on the server. It only has %s left."), NSSFormattedMemSize($freeSpace)).' '.$SYSADMIN;
            }
            if ( ! move_uploaded_file($_FILES[$key]['tmp_name'],$claimDir."/".$tmpname) ) {
              //  Exit gracefully -- dump database changes and remove the dropoff
              //  directory:
              $this->_dropbox->writeToLog("Error: failed to store dropoff files for $claimID");
              if ( ! rmdir_r($claimDir) ) {
                $this->_dropbox->writeToLog("Error: backing out new dropoff, unable to remove $claimDir - orphaned!");
              }
              if ( ! $this->_dropbox->database->DBRollbackTran() ) {
                $this->_dropbox->writeToLog("Error: backing out new dropoff, failed to ROLLBACK after botched dropoff $claimID, there may be orphan files");
              }
              // TRANSLATORS: %s = filename
              return sprintf(gettext("Trouble while attempting to drop '%s' into its drop-off directory."), htmlspecialchars($_FILES[$key]['name'])).' '.$SYSADMIN;
            } else {
              // Strip unwanted permissions - Want ug=r and nothing else
              chmod($claimDir."/".$tmpname, 0440);
            }

            // Checksum it if small enough (awkward to check size)
            $checksum = '';
            if ($checksumFiles) {
              $checksum = checksumFile($checksumcmd, $claimDir.'/'.$tmpname);
            }

            //  Add to database:
            if ( ! $this->_dropbox->database->DBAddFile1($dropoffID, $tmpname,
                             paramPrepare($_FILES[$key]['name']),
                             $bytes,
                             ( $_FILES[$key]['type'] ? $_FILES[$key]['type']
                               : "application/octet-stream" ),
                             paramPrepare($_POST["desc_".$i]),
                             $checksum) ) {
              //  Exit gracefully -- dump database changes and remove the dropoff
              //  directory:
              $this->_dropbox->writeToLog("Error: failed to add dropoff file to database for $claimID");
              if ( ! rmdir_r($claimDir) ) {
                $this->_dropbox->writeToLog("Error: backing out new dropoff, unable to remove $claimDir - orphaned!");
              }
              if ( ! $this->_dropbox->database->DBRollbackTran() ) {
                $this->_dropbox->writeToLog("Error: backing out new dropoff, failed to ROLLBACK after botched dropoff $claimID, there may be orphan files");
              }
              // TRANSLATORS: %s = filename
              return sprintf(gettext("Trouble while attempting to save the information for '%s'."), htmlspecialchars($_FILES[$key]['name'])).' '.$SYSADMIN;
            }
            
            //  That's right, one more file!
            $tplFiles[$realFileCount] = array();
            $tplFiles[$realFileCount]['name'] = paramPrepare($_FILES[$key]['name']);
            $tplFiles[$realFileCount]['type'] = $_FILES[$key]['type'];
            // Get filesize from local copy, not browser $tplFiles[$i]['size'] = NSSFormattedMemSize($_FILES[$key]['size']);
            $tplFiles[$realFileCount]['size'] = NSSFormattedMemSize($bytes);
            // Force the file description to max 100 chars
            $tplFiles[$realFileCount]['description'] =
              mb_strimwidth(paramPrepare($_POST["desc_".$i]), 0, 100, '...');
            if ($checksum != '')
              $tplFiles[$realFileCount]['checksum'] = $checksum;
            else
              $tplFiles[$realFileCount]['checksum'] = gettext('Not calculated');
            $realFileCount++;
          }
          $i++;
        }
        
        //  Once we get here, it's time to commit the stuff to the database:
        $this->_dropbox->database->DBCommitTran();

        $this->_dropoffID             = $dropoffID;
          
        //  At long last, fill-in the fields:
        $this->_claimID               = $claimID;
        $this->_claimPasscode         = $claimPasscode;
        $this->_claimDir              = $claimDir;
        
        $this->_authorizedUser        = $this->_dropbox->authorizedUser();
        
        $this->_note                  = $note;
        $this->_senderName            = $senderName;
        $this->_senderOrganization    = $senderOrganization;
        $this->_senderEmail           = $senderEmail;
        $this->_showIDPasscode        = $showIDPasscode;
        $senderIP                     = getClientIP();
        $this->_senderIP              = $senderIP;
        $senderHost = gethostbyaddr($senderIP);
        if ($senderHost != '') {
          $senderHost = '(' . $senderHost . ')';
        }
        $this->_confirmDelivery       = $confirmDelivery;
        $this->_informRecipients      = $informRecipients;
        $this->_informPasscode        = $informPasscode;
        $this->_created               = getdate();
        
        $this->_recipients            = $recipients;
        
        // This Drop-off request has been fulfilled, so kill the keys
        // to stop playback attacks.
        if (isset($req)) {
          $this->_dropbox->DeleteReqData($req);
        }
        if (isset($auth)) {
          $this->_dropbox->DeleteAuthData($auth);
        }

        // Work out the real email subject line.
        if ($reqSubject != '') {
          $emailSubject = $reqSubject;
        } else {
          if ($realFileCount == 1) {
            $emailSubject = sprintf($smarty->getConfigVars(
                                    'EmailSubjectTag') .
                                    // TRANSLATORS: email subject line
                                   gettext('%s has dropped off a file for you'),
                                    $senderName);
          } else {
            $emailSubject = sprintf($smarty->getConfigVars(
                                    'EmailSubjectTag') .
                                    // TRANSLATORS: email subject line
                                   gettext('%s has dropped off files for you'),
                                    $senderName);
          }
        }

        // How long till expiry?
        $daysLeft    = $this->_dropbox->retainDays();
        $timeLeft    = sprintf(gettext('%d days'), $daysLeft); // . ' ' . $smarty->getConfigVars('days');

        //  Construct the email notification and deliver:
        $smarty->assign('senderName',     $senderName);
        $smarty->assign('senderOrg',      $senderOrganization);
        $smarty->assign('senderEmail',    $senderEmail);
        $smarty->assign('showIP',         $this->_dropbox->emailSenderIP());
        $smarty->assign('senderIP',       $senderIP);
        $smarty->assign('senderHost',     $senderHost);
        $smarty->assign('note',           trim($note));
        $smarty->assign('subject',        $emailSubject);
        $smarty->assign('now',            timestampForTime(time()));
        $smarty->assign('claimID',        $claimID);
        $smarty->assign('claimPasscode',  $claimPasscode);
        $smarty->assign('informPasscode', $informPasscode);
        $smarty->assign('fileCount',      $realFileCount);
        $smarty->assign('retainDays',     $daysLeft);
        $smarty->assign('timeLeft',       $timeLeft);
        $smarty->assign('showIDPasscode', $showIDPasscode);
        $smarty->assignByRef('files',     $tplFiles);

        $emailTXTTemplate = $smarty->fetch('dropoff_email.tpl');
        $emailHTMLTemplate = '';
        if ($smarty->templateExists('dropoff_email_html.tpl')) {
          $emailHTMLTemplate = $smarty->fetch('dropoff_email_html.tpl');
        }

        // Update the address book entries for this user
        $this->_dropbox->updateAddressbook($recipients);

        // Inform all the recipients by email if they want me to
        if ($informRecipients) {
          // Do we want to Bcc the sender as well?
          // It also must be EITHER an internal user, OR they want
          // to Bcc external senders as well.
          $emailBcc = '';
          if ($this->_dropbox->bccSender() &&
              ($this->_dropbox->authorizedUser() ||
               $this->_dropbox->bccExternalSender() )) {
            $emailBcc = $senderEmail;
            // // and don't forget to encode it if there are intl chars in it
            // if (preg_match('/[^\x00-\x7f]/', $senderEmail)) {
            //   $emailBcc = "Bcc: =?UTF-8?B?".base64_encode(html_entity_decode($senderEmail))."?=".PHP_EOL;
            // } else {
            //   $emailBcc = "Bcc: $senderEmail".PHP_EOL;
            // }
          }
          // Make the mail come from the sender, not ZendTo
          foreach ( $recipients as $recipient ) {
            // In MyZendTo, don't send email to myself
            if ((preg_match('/^[yYtT1]/', MYZENDTO) && $senderEmail != $recipient[1]) || preg_match('/^[^yYtT1]/', MYZENDTO)) {
              $emailTXTContent = preg_replace('/__EMAILADDR__/', urlencode($recipient[1]), $emailTXTTemplate);
              $emailHTMLContent = preg_replace('/__EMAILADDR__/', urlencode($recipient[1]), $emailHTMLTemplate);
              $success = $this->_dropbox->deliverEmail(
                  array($recipient[1], $emailBcc),
                  $senderEmail,
                  $emailSubject,
                  $emailTXTContent,
                  $emailHTMLContent
               );
              $emailBcc = ''; // Only Bcc the sender on the first email out
              if ( ! $success ) {
                $this->_dropbox->writeToLog(sprintf("Error: failed to deliver notification email to %s for claimID $claimID",$recipient[1]));
              } else {
                $this->_dropbox->writeToLog(sprintf("Info: successfully delivered notification email to %s for claimID $claimID",$recipient[1]));
              }
            }
          }
        }
        
        //  Log our success:
        $this->_dropbox->writeToLog(sprintf("Info: new dropoff $claimID of %s created for $senderName <$senderEmail>",
               ( $realFileCount == 1 ? "1 file" : "$realFileCount files" )));
      } else {
        return gettext("Unable to add a drop-off record to the database.").' '.$SYSADMIN;
      }
    } else {
      return gettext("Unable to begin database transaction.").' '.$SYSADMIN;
    }
    return NULL;
  }

}

?>
